import { ɵdetectChanges, ɵrenderComponent, ViewEncapsulation } from '@angular/core';

function initProps(target) {
    target.schema.attrs.forEach((key)=>{
        Object.defineProperty(target, key, {
            get () {
                return target.component[key];
            },
            set (value) {
                target.component[key] = value;
            }
        });
    });
}
function initEvents(target) {
    const schema = target.schema;
    schema.events.forEach((outputEvent)=>{
        target.component[schema.outputs[outputEvent]].subscribe((info)=>{
            target.dispatchEvent(new CustomEvent(outputEvent, {
                detail: info
            }));
        });
    });
}

function createStyle(styles) {
    const style = new CSSStyleSheet();
    /// @ts-ignore
    style.replaceSync(styles.join('\n'));
    return style;
}
function createSchema(component, options) {
    const { inputs , styles , outputs  } = component['ɵcmp'];
    const directives = (options === null || options === void 0 ? void 0 : options.directives) ? options.directives.filter((t)=>t['ɵdir']
    ) : [];
    const components = (options === null || options === void 0 ? void 0 : options.directives) ? options.directives.filter((t)=>t['ɵcmp']
    ) : [];
    const pipes = (options === null || options === void 0 ? void 0 : options.pipes) ? options.pipes : [];
    const attrs = Object.keys(inputs);
    const events = Object.keys(outputs);
    const style = createStyle(styles);
    const props = new Map();
    attrs.forEach((attr)=>props.set(attr, component[attr])
    );
    return {
        attrs,
        events,
        props,
        style,
        directives,
        components,
        pipes,
        outputs
    };
}
const schema = new Map();

function autoChangeDetection(component, props, attrs) {
    attrs.forEach((key)=>{
        props.set(key, component[key]);
        Object.defineProperty(component, key, {
            get () {
                return props.get(key);
            },
            set (value) {
                props.set(key, value);
                ɵdetectChanges(component);
            }
        });
    });
    return component;
}

function renderNgComponent(componentType, options) {
    const type = componentType['ɵcmp'];
    const { directives , pipes , props , attrs  } = schema.get(type.id);
    type.directiveDefs = directives.map((d)=>d['ɵdir']
    );
    type.pipeDefs = pipes.map((p)=>p['ɵpipe']
    );
    const component = ɵrenderComponent(componentType, {
        ...options !== null && options !== void 0 ? options : {
        },
        rendererFactory: null
    });
    return autoChangeDetection(component, props, attrs);
}

class NgxElement extends HTMLElement {
    attributeChangedCallback(name, oldValue, newValue) {
        if (this.schema.attrs.includes(name)) {
            this[name] = newValue;
        }
    }
    constructor(type, options){
        super();
        const { id , encapsulation  } = type['ɵcmp'];
        this.schema = schema.get(id);
        const host = encapsulation === ViewEncapsulation.ShadowDom ? this.attachShadow({
            mode: 'open'
        }) : this;
        host.adoptedStyleSheets = [
            schema.get(id).style
        ];
        this.component = renderNgComponent(type, {
            ...options !== null && options !== void 0 ? options : {
            },
            host
        });
    }
}

function createCustomElement(componentType, options) {
    const type = componentType['ɵcmp'];
    const schemaResult = createSchema(componentType, options);
    const { components , directives , attrs  } = schemaResult;
    (options === null || options === void 0 ? void 0 : options.directives) && components.forEach((type1)=>renderCustomElement(type1, {
            ...options,
            directives
        })
    );
    schema.set(type.id, schemaResult);
    return (function() {
        class CustomElement extends NgxElement {
            static get observedAttributes() {
                return [
                    ...attrs
                ];
            }
            constructor(){
                super(componentType, options);
                initProps(this);
                initEvents(this);
            }
        }
        return CustomElement;
    })();
}
function renderCustomElement(componentType, options) {
    customElements.define(componentType['ɵcmp'].selectors[0][0], createCustomElement(componentType, options));
}

export { NgxElement, createCustomElement, renderCustomElement };
